import json

registered_transformation = {
    'int': lambda x: int(x),
    'str': lambda x: str(x),
    'float': lambda x: float(x),
    'list': lambda x: list(x),
    'json': lambda x: json.dumps(x),
    'json_dict': lambda x: json.loads(x),
}



import os
import sys
import uuid
from datetime import datetime
import unittest

from hypothesis import given
import hypothesis.strategies as st

from deepdiff import DeepDiff as ddiff

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))

from maat import validate

from maat import types

from maat import protected
from maat.exceptions import Invalid


def datetime_parse(val, key, formats="%Y-%m-%dT%H:%M:%S.%f", *args, **kwargs):
    """Parse datetime string 'val' in ISO format and return a datetime object,
    raise Invalid exception on error."""
    try:
        return datetime.strptime(val, formats)
    except Exception as e:
        raise Invalid(f'key: "{key}" contains invalid item "{type(val).__name__}": unable to convert from type')


class TestExtendMaat(unittest.TestCase):

    def test_extend_maat_datetime_happy_path(self):
        """Happy path test, Maat c"""
        
        types['custom_datetime'] = datetime_parse

        test_input = {
            "created": "2022-01-28T15:01:46.0000",
        }
        test_validation = {
            "created": {
                "type": "custom_datetime",
            }   
        }
        expected = {"created": datetime(2022, 1, 28, 15, 1, 46)}

        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_extend_maat_datetime_invalid_input(self):
        """Test with invalid time string, this string doens't adhere to full iso format.
        It only contains a date.
        """
        
        types['custom_datetime'] = datetime_parse

        test_input = {
            "created": "2022-01-28",
        }
        test_validation = {
            "created": {
                "type": "custom_datetime",
            }   
        }

        with self.assertRaisesRegex(Invalid, 'key: "created" contains invalid item "str": unable to convert from type'):
            _ = validate(test_input, test_validation)

    def test_extend_maat_datetime_with_argument(self):
        """Test with invalid time string, this string doens't adhere to full iso format.
        It only contains a date.
        """
        
        types['custom_datetime'] = datetime_parse

        test_input = {
            "created": "2022-01-28",
        }
        test_validation = {
            "created": {
                "type": "custom_datetime",
                "formats": "%Y-%m-%d",
            }   
        }

        expected = {"created": datetime(2022, 1, 28)}

        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_extend_maat_datetime_invalid_input_int(self):
        """Test with invalid time string, this string doens't adhere to full iso format.
        It only contains a date.
        """
        
        types['custom_datetime'] = datetime_parse

        test_input = {
            "created": 161234344,
        }
        test_validation = {
            "created": {
                "type": "custom_datetime",
                "formats": "%Y-%m-%d",
            }   
        }

        with self.assertRaisesRegex(Invalid, 'key: "created" contains invalid item "int": unable to convert from type'):
            _ = validate(test_input, test_validation)

if __name__ == "__main__":
    unittest.main()



from .validations import types
from .transformations import registered_transformation
from .exceptions import Invalid

NESTED = "nested"
OPTIONAL = "optional"
NULLABLE = "nullable"
DEFAULT = "default"
EMPTYLIST = "empty_list"
TRANSFORM = "transform"
PRETRANSFORM = "pre_transform"
SKIPFAILED = "skip_failed"

TYPE = "type"
TYPELIST = "list"
TYPEDICT = "dict"
TYPEASOARR = "aso_array"
TYPELISTDICTS = "list_dicts"

special_arguments = {NESTED, TYPELIST, TYPEASOARR, SKIPFAILED, NULLABLE, OPTIONAL, DEFAULT, PRETRANSFORM,
                     TRANSFORM, TYPELISTDICTS, EMPTYLIST, TYPEDICT, TYPE}


def get_transformation_func(item, type_transformation):
    """Check if transformation is set. If set try to find it in registerd transformation.
    If transformation is misspelled or sit or throw an exception"""
    transformation = item.get(type_transformation)
    if transformation is None:
        return lambda x: x
    try:
        return registered_transformation[transformation]
    except KeyError:
        raise Invalid(f"{transformation} is not registered as transformation")


def keys_equality(input_dict, counter_dict):
    try:
        return set(input_dict) <= set(counter_dict)
    except (TypeError, AttributeError):
        return False


def find_missing_keys(input_dict, counter_dict):
    try:
        found_keys = ", ".join(set(input_dict) - set(counter_dict))
        return f"{found_keys} :expected keys: {', '.join(counter_dict.keys())}"
    except (TypeError, AttributeError):
        raise Invalid(f"{input_dict} not a dictionary but is of type {type(input_dict).__name__}")


def scale(input_dict, counter_dict):  # noqa:C901

    if not keys_equality(input_dict, counter_dict):
        raise Invalid(f"invalid keys: {find_missing_keys(input_dict, counter_dict)}")

    validated_items = {}
    for key, item in counter_dict.items():
        try:
            val = input_dict[key]
        except KeyError:
            if DEFAULT in item:
                validated_items[key] = item[DEFAULT]
                continue
            elif item.get(OPTIONAL):
                continue
            else:
                raise Invalid(f'key:"{key}" is not set')

        # if the value is None check if value is allowed to be None
        if val is None and item.get(NULLABLE):
            validated_items[key] = None
            continue

        try:
            validation_func = types[item[TYPE]]
        except KeyError:
            raise Invalid(f"{item.get(TYPE)} is not registered as type")

        pre_transformation = get_transformation_func(item, PRETRANSFORM)
        post_transformation = get_transformation_func(item, TRANSFORM)

        # the validation can be done on top level, life is good
        if NESTED not in item:
            validated_items[key] = post_transformation(validation_func(key=key, val=pre_transformation(val), **item))
            continue

        type_ = item["type"]
        if type_ == TYPEDICT:
            validation_func(key=key, val=val, **item)
            validated_items[key] = post_transformation(scale(pre_transformation(input_dict[key]), counter_dict[key][NESTED]))

        elif type_ == TYPELIST:
            validation_func(key=key, val=val, **item)
            validated_items[key] = item.get(EMPTYLIST, []) if len(val) == 0 else []

            try:
                validation_func = types[item[NESTED][TYPE]]
            except KeyError:
                raise Invalid(f"{item[NESTED].get(TYPE)} is not registered as type")

            for nested_item in val:
                try:
                    validated_items[key].append(post_transformation(validation_func(key=key, val=pre_transformation(nested_item), **item[NESTED])))
                except Invalid:
                    if not item.get(SKIPFAILED):
                        raise

        elif type_ == TYPELISTDICTS:
            validation_func(key=key, val=val, **item)
            validated_items[key] = []
            for nested_item in val:
                try:
                    validated_items[key].append(post_transformation(scale(pre_transformation(nested_item), counter_dict[key][NESTED])))
                except Invalid:
                    if not item.get(SKIPFAILED):
                        raise

        elif type_ == TYPEASOARR:
            validation_func(key=key, val=val, **item)

            for nested_key, nested_val in val.items():
                # TODO this could be nicer
                if key not in validated_items:
                    validated_items[key] = {}
                if nested_key not in validated_items[key]:
                    validated_items[key][nested_key] = {}

                validated_items[key][nested_key] = validate(nested_val, counter_dict[key][NESTED])

        else:
            raise Invalid(f"type {type_} can't handle nested structures, use {TYPEASOARR}, {TYPELISTDICTS}, {TYPEDICT} instead")

    return validated_items


def validate(input_dict, counter_dict):  # noqa:C901

    if not keys_equality(input_dict, counter_dict):
        raise Invalid(f"invalid keys: {find_missing_keys(input_dict, counter_dict)}")

    validated_items = {}
    for key, item in counter_dict.items():
        try:
            val = input_dict[key]
        except KeyError:
            if DEFAULT in item:
                validated_items[key] = item[DEFAULT]
                continue
            elif item.get(OPTIONAL):
                continue
            else:
                raise Invalid(f'key:"{key}" is not set')

        # if the value is None check if value is allowed to be None
        if val is None and item.get(NULLABLE):
            validated_items[key] = None
            continue

        try:
            validation_func = types[item[TYPE]]
        except KeyError:
            raise Invalid(f"{item.get(TYPE)} is not registered as type")

        # the validation can be done on top level, life is good
        if NESTED not in item:
            validated_items[key] = validation_func(key=key, val=val, **item)
            continue

        type_ = item["type"]
        if type_ == TYPEDICT:
            validation_func(key=key, val=val, **item)
            validated_items[key] = validate(input_dict[key], counter_dict[key][NESTED])

        elif type_ == TYPELIST:
            validation_func(key=key, val=val, **item)
            validated_items[key] = item.get(EMPTYLIST, []) if len(val) == 0 else []

            try:
                validation_func = types[item[NESTED][TYPE]]
            except KeyError:
                raise Invalid(f"{item[NESTED].get(TYPE)} is not registered as type")

            for nested_item in val:
                try:
                    validated_items[key].append(validation_func(key=key, val=nested_item, **item[NESTED]))  # ADD test item nested
                except Invalid:
                    if not item.get(SKIPFAILED):
                        raise

        elif type_ == TYPELISTDICTS:
            validation_func(key=key, val=val, **item)
            validated_items[key] = []
            for nested_item in val:
                try:
                    validated_items[key].append(validate(nested_item, counter_dict[key][NESTED]))
                except Invalid:
                    if not item.get(SKIPFAILED):
                        raise

        elif type_ == TYPEASOARR:
            validation_func(key=key, val=val, **item)

            for nested_key, nested_val in val.items():
                # TODO this could be nicer
                if key not in validated_items:
                    validated_items[key] = {}
                if nested_key not in validated_items[key]:
                    validated_items[key][nested_key] = {}

                validated_items[key][nested_key] = validate(nested_val, counter_dict[key][NESTED])

        else:
            raise Invalid(f"type {type_} can't handle nested structures, use {TYPEASOARR}, {TYPELISTDICTS}, {TYPEDICT} instead")

    return validated_items



import os
import sys
import uuid
import unittest

from hypothesis import given
import hypothesis.strategies as st

from deepdiff import DeepDiff as ddiff

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from maat import validate

from maat import protected
from maat.exceptions import Invalid

from maat.validations import uuid_validation
from maat.validations import int_validation, str_validation, float_validation, list_validation, dict_validation


class TestValidation(unittest.TestCase):

    def setUp(self):
        self.test_input = {
            'id': 23,
            'name': 'John Doe',
            'type': 'banana',
        }
        self.test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'name': {'type': 'str', 'min_length': 1, 'max_length': 35, 'regex': r'(\w+ )(\w+)'},
            'type': {'type': 'str', 'choices': ['apple', 'banana', 'citrus']}
        }

    def test_validation(self):
        """Happy path test"""
        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)

        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validation_test_invalid_keys_exception(self):
        """Test invalid keys exception"""

        del self.test_input['type']

        with self.assertRaisesRegex(Invalid, 'key:"type" is not set'):
            _ = validate(self.test_input, self.test_validation)

    def test_validation_test_remove_key_and_set_optional(self):
        """Test remove key and set optional"""

        del self.test_input['type']
        self.test_validation['type']['optional'] = True

        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {})

    def test_validation_test_remove_key_and_remove_required(self):
        """Test remove key and set optional"""

        del self.test_input['type']
        self.test_validation['type']['optional'] = True
        excepted_value = 'banana'
        self.test_validation['type']['default'] = excepted_value

        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {'dictionary_item_removed': set(["root['type']"])})
        self.assertEqual(validated_items['type'], excepted_value)

    def test_validation_test_remove_key_and_set_default(self):
        """Test remove key and set optional"""

        del self.test_input['type']
        excepted_value = 'banana'
        self.test_validation['type']['default'] = excepted_value

        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {'dictionary_item_removed': set(["root['type']"])})
        self.assertEqual(validated_items['type'], excepted_value)

    def test_validate_invalid_id(self):
        """Set id to invalid value, expect Exception"""
        self.test_input['id'] = -1

        with self.assertRaisesRegex(Invalid, 'key: "id" contains invalid item "-1": integer is less then 1'):
            _ = validate(self.test_input, self.test_validation)

    def test_validate_invalid_int_value_name(self):
        self.test_input['name'] = 30
        with self.assertRaisesRegex(Invalid, 'key: "name" contains invalid item "30" with type "int": not of type string'):
            _ = validate(self.test_input, self.test_validation)

        # test validator of name to make the previous invalid value valid.
        self.test_validation['name'] = {'type': 'int'}
        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {})

    def test_validate_float_value_name(self):
        self.test_input['name'] = 30.0
        with self.assertRaisesRegex(Invalid, 'key: "name" contains invalid item "30.0" with type "float": not of type string'):
            _ = validate(self.test_input, self.test_validation)

        # test validator of name to make the previous invalid value valid.
        self.test_validation['name'] = {'type': 'float'}
        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {})

    def test_validate_programming_error_trying_to_validate_non_dict(self):
        test_input = [1, 2, 4]
        with self.assertRaisesRegex(Invalid, "\[1, 2, 4\] not a dictionary but is of type list"):
            _ = validate(test_input, self.test_validation)

        test_input = set([1, 2, 3])
        with self.assertRaisesRegex(Invalid, "\{1, 2, 3\} not a dictionary but is of type set"):
            _ = validate(test_input, self.test_validation)

        test_input = 1
        with self.assertRaisesRegex(Invalid, "1 not a dictionary but is of type int"):
            _ = validate(test_input, self.test_validation)

        test_input = None
        with self.assertRaisesRegex(Invalid, "None not a dictionary but is of type None"):
            _ = validate(test_input, self.test_validation)

    def test_validate_nested_dict(self):
        self.test_input = {
            'id': 23,
            'addresses': {
                'street': 'John Doe Street',
                'number': 123,
            }
        }
        self.test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'dict', 'key_regex': r'\w+', 'nested': {
                'street': {'type': 'str', 'min_length': 5, 'max_length': 99},
                'number': {'type': 'int', 'min_amount': 1},
                }
            }
        }
        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {})

    def test_validate_nested_dict_fail_on_regex(self):
        self.test_input = {
            'id': 23,
            'addresses': {
                'street': 'John Doe Street',
                'number': 123,
            }
        }
        self.test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'dict', 'key_regex': r'\d+', 'nested': {
                'street': {'type': 'str', 'min_length': 5, 'max_length': 99},
                'number': {'type': 'int', 'min_amount': 1},
                }
            }
        }
        with self.assertRaisesRegex(Invalid, 'addresses: has dictionary key "street" that does not adhere to regex \\\d+'):
            _ = validate(self.test_input, self.test_validation)

    def test_validate_nested_list(self):
        self.test_input = {
            'id': 23,
            'addresses': [
                {'street': 'John Doe Street'},
                {'street': 'John Doe Street'},
                {'street': 'John Doe Street'},
                {'street': 'John Doe Street'},
            ]
        }
        self.test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'list_dicts',  'nested': {
                    'street': {'type': 'str', 'min_length': 5, 'max_length': 99},
                }
            }
        }

        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {})

    def test_validate_two_deep_nested_dict(self):
        self.test_input = {
            'id': 23,
            'addresses': {
                'street': {
                    'two': 'deep',
                }
            }
        }
        self.test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'dict', 'key_regex': r'\w+',
                'nested': {
                    'street': {'type': 'dict', 'min_amount': 1, 'max_length': 99,
                        'nested': {
                            'two': {'type': 'str', 'min_length': 3, 'max_length': 99},
                        }
                    }
                }
            }
        }
        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)
        self.assertEqual(difference, {})

    def test_validate_very_nested_dict(self):
        nested_dict = {
            'data': {
                'people': {
                    '7': {
                        'id': 7,
                        'name': 'John Doe',
                        'type': 'mimic',
                        'x': 823.6228647149701,
                        'y': 157.57736006592654,
                        'address': {
                            'id': 23,
                            'addresses': {
                                'street': {
                                    'two': 'deep',
                                    '222': 'deep',
                                }
                            }
                        }
                    },
                    '208': {
                        'id': 208,
                        'name': 'John Doe Too',
                        'type': 'person',
                        'x': 434.9446032612515,
                        'y': 580.0,
                        'address': {
                            'id': 23,
                            'addresses': {
                                'street': {
                                    'two': 'deep',
                                    '222': 'deep',
                                }
                            }
                        }
                    }
                },
                'streets': {
                    'id': 23,
                    'addresses': [
                        {'street': 'John Doe Street'},
                        {'street': 'John Doe Street'},
                        {'street': 'John Doe Street'},
                        {'street': 'John Doe Street'},
                    ]
                }
            }
        }
        addresses_item = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'dict', 'nested': {
                'street': {'type': 'dict', 'min_amount': 1, 'max_length': 99, 'nested': {
                    'two': {'type': 'str', 'min_length': 3, 'max_length': 99},
                    '222': {'type': 'str', 'min_length': 3, 'max_length': 99},
                        }
                    }
                }
            }
        }

        geo_item = {
            'id': {'type': 'int', 'min_amount': 1},
            'name': {'type': 'str', 'min_length': 1, 'max_length': 35, 'regex': '([^\s]+)'},
            'type': {'type': 'str', 'min_length': 1, 'max_length': 25, 'regex': r'([^\s]+)'},
            'x': {'type': 'float'},
            'y': {'type': 'float'},
            'address': {'type': 'dict',
                'nested': addresses_item}
        }

        nested_dict_validation = {
            'data': {'type': 'dict', 'nested': {
                'people': {'type': 'aso_array', 'min_amount': 1, 'max_amount': 99,
                    'nested': geo_item},
                'streets': {'type': 'dict', 'nested': {
                    'id': {'type': 'int', 'min_amount': 1},
                    'addresses': {'type': 'list_dicts', 'nested': {
                        'street': {'type': 'str', 'min_length': 1, 'max_length': 99}
                                }
                            }
                        }
                    }
                }
            }
        }
        validated_items = validate(nested_dict, nested_dict_validation)
        difference = ddiff(validated_items, nested_dict)
        self.assertEqual(difference, {})


    def test_validate_very_nested_dict_fail_lowest_item(self):
        nested_dict = {
            'data': {
                'people': {
                    '7': {
                        'id': 7,
                        'name': 'John Doe',
                        'type': 'mimic',
                        'x': 823.6228647149701,
                        'y': 157.57736006592654,
                        'address': {
                            'id': 23,
                            'addresses': {
                                'street': {
                                    'two': 'deep',
                                    '222': 'deep',
                                }
                            }
                        }
                    },
                    '208': {
                        'id': 208,
                        'name': 'John Doe Too',
                        'type': 'person',
                        'x': 434.9446032612515,
                        'y': 580.0,
                        'address': {
                            'id': 23,
                            'addresses': {
                                'street': {
                                    'two': 'deep',
                                    '222': 'deep',
                                }
                            }
                        }
                    }
                },
                'streets': {
                    'id': 23,
                    'addresses': [
                        {'street': 'John Doe Street'},
                        {'street': 'John Doe Street'},
                        {'street': 'John Doe Street'},
                        {'street': 'John Doe Street'},
                    ]
                }
            }
        }
        addresses_item = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'dict', 'nested': {
                'street': {'type': 'dict', 'min_amount': 1, 'max_length': 99, 'nested': {
                    'two': {'type': 'str', 'min_length': 3, 'max_length': 99},
                    '222': {'type': 'str', 'min_length': 3, 'max_length': 99, 'choices': [
                        'not_part_of_choices',
                        'fail_here']},
                        }
                    }
                }
            }
        }

        geo_item = {
            'id': {'type': 'int', 'min_amount': 1},
            'name': {'type': 'str', 'min_length': 1, 'max_length': 35, 'regex': '([^\s]+)'},
            'type': {'type': 'str', 'min_length': 1, 'max_length': 25, 'regex': r'([^\s]+)'},
            'x': {'type': 'float'},
            'y': {'type': 'float'},
            'address': {'type': 'dict',
                'nested': addresses_item}
        }

        nested_dict_validation = {
            'data': {'type': 'dict', 'nested': {
                'people': {'type': 'aso_array', 'min_amount': 1, 'max_amount': 99,
                    'nested': geo_item},
                'streets': {'type': 'dict', 'nested': {
                    'id': {'type': 'int', 'min_amount': 1},
                    'addresses': {'type': 'dict', 'nested': {
                        'street': {'type': 'str', 'min_length': 1, 'max_length': 99}
                                }
                            }
                        }
                    }
                }
            }
        }
        exp_exc_msg = "key: \"222\" contains invalid item \"deep\": not in valid choices \['not_part_of_choices', 'fail_here'\]"
        with self.assertRaisesRegex(Invalid, exp_exc_msg):
            _ = validate(nested_dict, nested_dict_validation)

    def test_validate_item_200_deep(self):
        """Lower depth for deepdiff limits, then later tests"""
        input_dict = current = {}
        counter_dict = counter_current = {}

        times = 200
        for _ in range(times):
            current['nested_dic'] = {}
            current = current['nested_dic']

        # set last item
        current['last'] = 4

        counter_current['nested_dic'] = {'type': 'dict', 'min_amount': 0, 'max_amount': 10, 'nested': {}}
        for _ in range(times - 1):
            counter_current['nested_dic']['nested'] = {'nested_dic': {'type': 'dict', 'min_amount': 0, 'max_amount': 10, 'nested': {}}}
            counter_current = counter_current['nested_dic']['nested']

        counter_current['nested_dic']['nested'] = {'last': {'type': 'int', 'min_amount': 1, 'max_amount': 10}}

        validated_items = validate(input_dict, counter_dict)
        difference = ddiff(validated_items, input_dict)
        self.assertEqual(difference, {})

    def test_validate_item_800_deep_invalid_item(self):
        input_dict = current = {}
        counter_dict = counter_current = {}

        times = 800
        for _ in range(times):
            current['nested_dic'] = {}
            current = current['nested_dic']

        # set last item
        current['last'] = 4

        counter_current['nested_dic'] = {'type': 'dict', 'min_amount': 0, 'max_amount': 10, 'nested': {}}
        for _ in range(times - 1):
            counter_current['nested_dic']['nested'] = {'nested_dic': {'type': 'dict', 'min_amount': 0, 'max_amount': 10, 'nested': {}}}
            counter_current = counter_current['nested_dic']['nested']

        # increase minimal amount to 5
        counter_current['nested_dic']['nested'] = {'last': {'type': 'int', 'min_amount': 5, 'max_amount': 10}}

        with self.assertRaisesRegex(Invalid, 'key: "last" contains invalid item "4": integer is less then 5'):
            _ = validate(input_dict, counter_dict)


class TestValidatorPropertyBased(unittest.TestCase):

    def result_as_bool(self, func, *args, **kwargs):
        try:
            func(*args, **kwargs)
        except Invalid:
            return False

        return True

    @given(val=st.text(), min_length=st.integers(), max_length=st.integers())
    def test_string_size_validation_test(self, val, min_length, max_length):
        expected = (max_length +1) > len(val) > (min_length -1)
        result = self.result_as_bool(str_validation, val=val, min_length=min_length, max_length=max_length)
        self.assertEqual(expected, result)

    @given(val=st.integers(), min_amount=st.integers(),
           max_amount=st.integers())
    def test_int_validation_min_max(self, val, min_amount, max_amount):
        expected = val <= max_amount and val >= min_amount
        result = self.result_as_bool(int_validation, val=val, min_amount=min_amount, max_amount=max_amount)

        self.assertEqual(expected, result)

    #TODO Python long ints to float conversion goes wrong, thus all tests too.
    @given(val=st.floats(), min_amount=st.integers(), max_amount=st.integers())
    def test_float_validation_min_max(self, val, min_amount, max_amount):
        expected = (val <= max_amount and val >= min_amount) or (val == max_amount and min_amount == max_amount)
        result = self.result_as_bool(float_validation, val=val, min_amount=min_amount, max_amount=max_amount)
        self.assertEqual(expected, result)


class ValidatorTests(unittest.TestCase):

    def test_str_validation_untouched_test(self):
        expected = 'banana'
        result = str_validation(val=expected)
        self.assertEqual(expected, result)

    def test_str_validation_min_length(self):
        expected = '1234'
        result = str_validation(val=expected, min_length=4)
        self.assertEqual(expected, result)

    def test_str_validation_min_length_failed(self):
        expected = '123'
        with self.assertRaises(Invalid):
            _ = str_validation(val=expected, min_length=4)

    def test_validation_max_length(self):
        expected = '1234'
        result = str_validation(val=expected, max_length=4)
        self.assertEqual(expected, result)

    def test_str_validation_max_length_failed(self):
        expected = '12345'
        with self.assertRaises(Invalid):
            _ = str_validation(val=expected, max_length=4)

    def test_str_validation_reqex_test(self):
        """Match only two words with regex"""
        regex = r'^(\w+ )(\w+)$'
        expected = 'apple tree'
        result = str_validation(val=expected, regex=regex)
        self.assertEqual(expected, result)

    def test_str_validation_failed_regex_test(self):
        """Match only two words with regex"""
        regex = r'^(\w+ )(\w+)$'
        expected = 'apple tree barks what is '
        with self.assertRaises(Invalid):
            _ = str_validation(key=None, val=expected, regex=regex)

    def test_str_validation_choices_test(self):
        expected = ['apple', 'banana', 'citrus']
        for item in expected:
            result = str_validation(val=item, choices=expected)
            self.assertEqual(item, result)

    def test_str_validation_choices_wrong_item_test(self):
        choices = ['apple', 'banana', 'citrus']
        wrong_choices = ['star', 'moon']
        for item in wrong_choices:
            with self.assertRaises(Invalid):
                _ = str_validation(val=item, choices=choices)

    def test_int_validation_convert_to_int_test(self):
        expected = '5'
        with self.assertRaises(Invalid):
            _ = int_validation(val=expected)

        result = int_validation(val=expected, cast=True)
        self.assertEqual(int(expected), result)

    def test_int_validation_convert_fail_test(self):
        expected = 'appel'
        with self.assertRaises(Invalid):
            _ = int_validation(val=expected)

        with self.assertRaises(Invalid):
            _ = int_validation(val=expected, cast=True)

    def test_float_validation_convert_to_float_test(self):
        expected = '5'
        with self.assertRaises(Invalid):
            _ = float_validation(val=expected)

        result = float_validation(val=expected, cast=True)
        self.assertEqual(float(expected), result)

    def test_float_validation_convert_fail_test(self):
        expected = 'appel'
        with self.assertRaises(Invalid):
            _ = float_validation(val=expected)

        with self.assertRaises(Invalid):
            _ = float_validation(val=expected, cast=True)

    def test_list_validation_invalid_val_test(self):
        non_list_types = ['str', 2, {}, tuple(), float()]
        for item in non_list_types:
            with self.assertRaises(Invalid):
                _ = list_validation(val=item)

    def test_list_validation_invalid_too_long_test(self):
        too_long = [1, 2, 3, 4, 5]
        with self.assertRaises(Invalid):
            _ = list_validation(val=too_long, max_amount=4)

    def test_list_validation_invalid_too_short_test(self):
        too_short = [1, 2, 3, 4, 5]
        with self.assertRaises(Invalid):
            _ = list_validation(val=too_short, min_amount=6)

    def test_dict_validation_invalid_val_test(self):
        non_dict_types = ['str', 2, [], tuple(), float()]
        for item in non_dict_types:
            with self.assertRaises(Invalid):
                _ = dict_validation(val=item)

    def test_dict_validation_invalid_too_long_test(self):
        too_long = {i: i for i in range(5)}
        with self.assertRaises(Invalid):
            _ = dict_validation(val=too_long, max_amount=4)

    def test_dict_validation_invalid_too_short_test(self):
        too_short = {i: i for i in range(5)}
        with self.assertRaises(Invalid):
            _ = dict_validation(val=too_short, min_amount=6)

    def test_dict_validation_valid_regex_test(self):
        dict_with_valid_keys = {str(i): i for i in range(5)}
        result = dict_validation(val=dict_with_valid_keys, key_regex='[0-9]')
        difference = ddiff(dict_with_valid_keys, result)
        self.assertEqual(difference, {})

    def test_dict_validation_invalid_regex_test(self):
        dict_with_invalid_keys = {str(i): i for i in range(5)}
        dict_with_invalid_keys['a'] = 'not valid'
        with self.assertRaises(Invalid):
            _ = dict_validation(val=dict_with_invalid_keys, key_regex='[0-9]')

    def test_dict_validation_invalid_types_regex_test(self):
        dict_with_invalid_keys = {
            'keys': None,
            None: None,
            1: None,
        }
        with self.assertRaises(Invalid):
            _ = dict_validation(val=dict_with_invalid_keys, key_regex='[0-9]')

    def test_uuid_validation_valid_uuids(self):
        for _ in range(10):
            expected = str(uuid.uuid4())
            result = uuid_validation(val=expected)
            self.assertEqual(expected, result)

    def test_uuid_validation_invalid_uuids(self):
        none_uuids = [
            str(uuid.uuid4()) + 'a',
            'appels',
            '92ue92ueof-03i903ir039-034i309',
            None,
            [],
            {},
        ]
        for item in none_uuids:
            with self.assertRaises(Invalid):
                _ = uuid_validation(val=item)

class ValidatorWrongInputTests(unittest.TestCase):
    def setUp(self):
        self.test_input = {
            'id': 23,
            'name': 'John Doe',
            'type': 'banana',
        }
        self.test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'name': {'type': 'str', 'min_length': 1, 'max_length': 35, 'regex': r'(\w+ )(\w+)'},
            'type': {'type': 'str', 'choices': ['apple', 'banana', 'citrus']}
        }

    def test_missing_key_validation(self):
        """Items that will raise an exception"""
        del(self.test_validation['name'])

        expected_expection_msg = 'invalid keys: name :expected keys: id, type'
        if sys.version_info[1] < 6:
            expected_expection_msg = 'invalid keys: name'

        with self.assertRaisesRegex(Invalid, expected_expection_msg):
            _ = validate(self.test_input, self.test_validation)


    def test_wrong_key_validation(self):
        """Items that will raise an exception"""
        self.test_validation['user'] = self.test_validation['name']
        del(self.test_validation['name'])

        expected_expection_msg = 'invalid keys: name :expected keys: id, type, user'
        if sys.version_info[1] < 6:
            expected_expection_msg = 'invalid keys: name'

        with self.assertRaisesRegex(Invalid, expected_expection_msg):
            _ = validate(self.test_input, self.test_validation)

    def test_missing_input_key_validation(self):
        """Items that will raise an exception"""
        del(self.test_input['id'])

        with self.assertRaisesRegex(Invalid, 'key:"id" is not set'):
            _ = validate(self.test_input, self.test_validation)

    def test_not_set_type(self):
        """Approriate message shown when trying to use a type that is not registered"""
        test_input = {'id': 23}
        test_validation = {'id': {'type': 'integer'}}

        with self.assertRaisesRegex(Invalid, 'integer is not registered as type'):
            _ = validate(test_input, test_validation)

    def test_not_set_type_nested(self):
        """Approriate message shown when trying to use a type that is not registered"""
        from maat import types
        types['newtype'] = lambda a: a

        test_input = {'id': 23}
        test_validation = {'id': {'type': 'newtype', 'nested': {'type': 'int'}}}

        with self.assertRaisesRegex(Invalid, "type newtype can't handle nested structures, use aso_array, list_dicts, dict instead"):
            _ = validate(test_input, test_validation)
        del types['newtype']

    def test_item_nullable(self):
        test_input = {'id': None}
        test_validation = {'id': {'type': 'float'}}
        with self.assertRaisesRegex(Invalid, 'key: \"id\" contains invalid item \"None\" with type \"NoneType\": not of type float'):
            _ = validate(test_input, test_validation)

        test_validation = {'id': {'type': 'float', 'nullable': True}}
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, test_input)
        self.assertEqual(difference, {})


    def test_validate_skip_instead_of_fail_within_nested_list_with_custom_validation(self):
        from maat import types
        def denylist_address(val, key, *args, **kwargs):
            """Since this is a example denylisted is hardcoded.
            It could come from config or an key value store at runtime
            Either within this function or given as argument.
            """
            denylist = ['denylisted', 'deny_listed', 'DENYLISTED']
            if val in denylist:
                raise Invalid('{0} is in denylist of {1}'.format(key, val))
            return val

        types['valid_address'] = denylist_address
        test_input = {
            'id': 23,
            'addresses': [
                'denylisted',
                'valid adress',
                'also valid address',
                'deny_listed',
                'valid again',
                'DENYLISTED',
            ]
        }
        test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'list', 'skip_failed': True, 'nested': {
                'type': 'valid_address'
                }
            }
        }
        expected_result = {
            'id': 23,
            'addresses': [
                'valid adress',
                'also valid address',
                'valid again',
            ]
        }
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected_result)
        self.assertEqual(difference, {})

    def test_validate_custom_validation_context_list_type_is_not_set(self):
        test_input = {
            'id': 23,
            'addresses': [
                'denylisted',
                'valid adress',
                'also valid address',
                'deny_listed',
                'valid again',
                'DENYLISTED',
            ]
        }
        test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'list', 'nested': {
                'type': 'not_registered'
                }
            }
        }

        with self.assertRaisesRegex(Invalid, 'not_registered is not registered as type'):
            _ = validate(test_input, test_validation)

    def test_validate_fail_within_nested_list_with_custom_validation(self):
        from maat import types
        def denylist_address(val, key, *args, **kwargs):
            """Since this is a example denylisted is hardcoded.
            It could come from config or an key value store at runtime
            Either within this function or given as argument.
            """
            denylist = ['denylisted', 'deny_listed', 'DENYLISTED']
            if val in denylist:
                raise Invalid('{0} is in denylist of {1}'.format(key, val))
            return val

        types['valid_address'] = denylist_address
        test_input = {
            'id': 23,
            'addresses': [
                'valid adress',
                'denylisted',
            ]
        }
        test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'list', 'nested': {
                'type': 'valid_address'
                }
            }
        }
        with self.assertRaisesRegex(Invalid, "addresses is in denylist of denylisted"):
            _ = validate(test_input, test_validation)

    def test_validate_fail_within_nested_list_dicts_with_custom_validation(self):
        from maat import types
        def denylist_address(val, key, *args, **kwargs):
            """Since this is a example denylisted is hardcoded.
            It could come from config or an key value store at runtime
            Either within this function or given as argument.
            """
            denylist = ['denylisted', 'deny_listed', 'DENYLISTED']
            if val in denylist:
                raise Invalid('{0} is in denylist of {1}'.format(key, val))
            return val

        types['valid_address'] = denylist_address
        test_input = {
            'id': 23,
            'addresses': [
                {'street': 'valid adress'},
                {'street': 'denylisted'},
            ]
        }
        test_validation = {
            'id': {'type': 'int', 'min_amount': 1},
            'addresses': {'type': 'list_dicts', 'nested': {
              'street': {'type': 'valid_address'}}
            }
        }
        with self.assertRaisesRegex(Invalid, "street is in denylist of denylisted"):
            _ = validate(test_input, test_validation)

class TestValidationDecorator(unittest.TestCase):

    def setUp(self):
        self.test_input = {
            'number': 23,
            'name': 'John Doe',
            'kind': 'banana',
        }
        self.test_validation = {
            'number': {'type': 'int', 'min_amount': 1},
            'name': {'type': 'str', 'min_length': 1, 'max_length': 35, 'regex': r'(\w+ )(\w+)'},
            'kind': {'type': 'str', 'choices': ['apple', 'banana', 'citrus']}
        }

    def test_validation_of_arguments(self):
        """Happy path test"""
        @protected(self.test_validation)
        def foo(number, name, kind):
            return locals()
        result = foo(**self.test_input)
        difference = ddiff(result, self.test_input)

        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validation_of_arguments_diffirent_input(self):

        @protected(self.test_validation)
        def foo(number, name, kind):
            return locals()

        number = self.test_input['number']
        name = self.test_input['name']
        kind = self.test_input['kind']

        # all arguments are part of kwargs
        result = foo(number=number, name=name, kind=kind)
        difference = ddiff(result, self.test_input)
        self.assertEqual(difference, {})

        # all arguments all part of args
        result = foo(number, name, kind)
        difference = ddiff(result, self.test_input)
        self.assertEqual(difference, {})

        # mixed arg, kwargs arguments
        result = foo(number, name, kind=kind)
        difference = ddiff(result, self.test_input)
        self.assertEqual(difference, {})

    def test_validation_of_argument_fail(self):
        """Test with validation failures"""

        @protected(self.test_validation)
        def foo(number, name, kind):
            return locals()

        # change type of number from int to str
        with self.assertRaisesRegex(Invalid, 'key: "number" contains invalid item "2" with type "str": not of type int'):
            result = foo(number='2', name='foo bar', kind='apple')

        # let's remove an argument
        with self.assertRaisesRegex(Invalid, 'key:"kind" is not set'):
            result = foo(number=2, name='foo bar')

        # tests that functions still works
        result = foo(**self.test_input)
        difference = ddiff(result, self.test_input)

        self.assertEqual(difference, {})

    def test_validation_of_argument_fail_returns_none(self):
        """Test with validation failures handle them and return None"""

        @protected(self.test_validation, fail_is_none=True)
        def foo(number, name, kind):
            return locals()

        # change type of number from int to str
        result = foo(number='2', name='foo bar', kind='apple')
        self.assertEqual(result, None)
        # let's remove an argument
        result = foo(number=2, name='foo bar')
        self.assertEqual(result, None)

        # tests that functions still works
        result = foo(**self.test_input)
        difference = ddiff(result, self.test_input)

        self.assertEqual(difference, {})

    def test_validation_of_argument_fail_with_custom_exception(self):
        """Test with validation failures raises an custom exception"""

        @protected(self.test_validation, custom_exception=KeyError)
        def foo(number, name, kind):
            return locals()

        # change type of number from int to str
        with self.assertRaisesRegex(KeyError, ''):
            result = foo(number='2', name='foo bar', kind='apple')

        # let's remove an argument
        with self.assertRaisesRegex(KeyError, ''):
            result = foo(number=2, name='foo bar')

        # tests that functions still works
        result = foo(**self.test_input)
        difference = ddiff(result, self.test_input)

        self.assertEqual(difference, {})


if __name__ == '__main__':
    unittest.main()



import os
import sys
import uuid
import unittest

from deepdiff import DeepDiff as ddiff

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))
import maat

from maat.exceptions import Invalid

from maat.validations import str_validation, float_validation, list_validation, dict_validation


class Foo(str):
    """Error once when str is called.
    Test report should not have Exception
    """

    def __init__(self):
        self.first_time = True

    def __str__(self):
        if self.first_time:
            self.first_time = False
            raise ValueError
        return "i'm foo who are you?"


class TestHandleStrCast(unittest.TestCase):

    def setUp(self):
        pass

    def test_validation(self):
        """Happy path test"""

        expected = Foo()
        result = str_validation(val=expected)
        self.assertEqual(result, expected)


    def test_validation_with_cast(self):
        
        with self.assertRaisesRegex(Invalid, 'key: "None" contains invalid item "Foo": unable to convert from type  to str'):
            str_validation(val=Foo(), cast=True)


if __name__ == '__main__':
    unittest.main()



import os
import sys
import unittest

from deepdiff import DeepDiff as ddiff

import base64
from Crypto.Cipher import AES

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))


def encode(msg_text):
    """
    Encrypts a message using AES ECB mode with right-justified padding and encodes the result in base64.

    Args:
        msg_text (str): The plaintext message to be encrypted.

    Returns:
        str: The base64-encoded encrypted message.
    """
    secret_key = os.environ.get('secret', 'veryMuchSecret').encode('utf-8')
    msg_text_bytes = msg_text.rjust(32).encode('utf-8')
    cipher = AES.new(secret_key, AES.MODE_ECB)
    encrypted_bytes = cipher.encrypt(msg_text_bytes)

    return base64.b64encode(encrypted_bytes).decode('utf-8')


def decode(encoded):
    """
    Decrypts a base64-encoded message encrypted using AES ECB mode and removes leading spaces.

    Args:
        encoded (str): The base64-encoded encrypted message.

    Returns:
        str: The decrypted message with leading spaces removed.
    """
    secret_key = os.environ.get('secret', 'veryMuchSecret').encode('utf-8')

    cipher = AES.new(secret_key, AES.MODE_ECB)
    encrypted_bytes = base64.b64decode(encoded)
    decrypted_bytes = cipher.decrypt(encrypted_bytes)
    decrypted_text = decrypted_bytes.decode('utf-8')

    return decrypted_text.lstrip()


import maat  # noqa:E402
os.environ['secret'] = 'super super secret key'.rjust(32)
maat.registered_transformation['encode'] = encode
maat.registered_transformation['decode'] = decode


class TestEncryptDecrypt(unittest.TestCase):
    """For eu compliance reasons data user data is stored encrypted in
    the database. Following examples take the stated above as true.
    Don't use the encryption that is used in this example in production."""
    def setUp(self):
        self.expected = {
            'name': 'John Doe',
            'address': 'John Doe Street',
        }

        self.expected_encoded = {
            'name': 'LcyInWDZsUv22ocRHM3+yg7QO9ArjlhP2R9v5CSZIRc=',
            'address': 'LcyInWDZsUv22ocRHM3+yryn2OYg2jesvpgClxA/sdQ=',
        }

    def test_validate_and_transform_incoming_data(self):
        """This test takes data, validates and then encrypt"""
        test_input = {
            'name': 'John Doe',
            'address': 'John Doe Street',
        }
        counter_dict = {
            'name': {
                'type': 'str', 'regex': 'John Doe', 'transform': 'encode'
            },
            'address': {
                'type': 'str', 'regex': 'John Doe Street', 'transform': 'encode'
            },
        }
        validated_items = maat.scale(test_input, counter_dict)
        difference = ddiff(validated_items, self.expected_encoded)

        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_decrypt_validate_and_encrypt(self):
        """This test takes encrypted data, validates and then encrypt"""
        test_input = {
            'name': 'LcyInWDZsUv22ocRHM3+yg7QO9ArjlhP2R9v5CSZIRc=',
            'address': 'LcyInWDZsUv22ocRHM3+yryn2OYg2jesvpgClxA/sdQ=',
        }

        counter_dict = {
            'name': {
                'type': 'str', 'regex': 'John Doe',
                'pre_transform': 'decode', 'transform': 'encode'
            },
            'address': {
                'type': 'str', 'regex': 'John Doe Street',
                'pre_transform': 'decode', 'transform': 'encode'
            },
        }
        validated_items = maat.scale(test_input, counter_dict)
        difference = ddiff(validated_items, self.expected_encoded)

        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_decrypt_validate_and_encrypt_fail_on_name(self):
        """This test takes encrypted data, validates and before
        encypting fails because the value of name is not valid"""
        test_input = {
            'name': 'LcyInWDZsUv22ocRHM3+yg7QO9ArjlhP2R9v5CSZIRc=',
            'address': 'LcyInWDZsUv22ocRHM3+yryn2OYg2jesvpgClxA/sdQ=',
        }

        counter_dict = {
            'name': {
                'type': 'str', 'regex': 'Jane',
                'pre_transform': 'decode', 'transform': 'encode'
            },
            'address': {
                'type': 'str', 'regex': 'John Doe Street',
                'pre_transform': 'decode', 'transform': 'encode'
            },
        }

        err_msg = 'key: "name" contains invalid item "John Doe": does not adhere to regex rules Jane'
        with self.assertRaisesRegex(maat.Invalid, err_msg):
            _ = maat.scale(test_input, counter_dict)


if __name__ == '__main__':
    unittest.main()



import re
import math

from uuid import UUID

from .exceptions import Invalid


def str_validation(val, key=None, min_length=None, max_length=None, regex=None, choices=None, cast=None, *args, **kwargs):
    if cast:
        try:
            val = str(val)
        except (ValueError, TypeError):
            raise Invalid(f'key: "{key}" contains invalid item "{type(val).__name__}": unable to convert from type  to str')

    if not isinstance(val, str):
        raise Invalid(f'key: "{key}" contains invalid item "{val}" with type "{type(val).__name__}": not of type string')

    if min_length is not None and len(val) < min_length:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": less then minimal length of {min_length}')

    if max_length is not None and len(val) > max_length:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": more then maximal length of {max_length}')

    if regex is not None and not bool(re.match(regex, val)):
        raise Invalid(f'key: "{key}" contains invalid item "{val}": does not adhere to regex rules {regex}')

    if choices is not None and val not in choices:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": not in valid choices {choices}')

    return val


def int_validation(val, key=None, min_amount=None, max_amount=None, cast=None, *args, **kwargs):
    if cast:
        try:
            val = int(val)
        except (ValueError, TypeError):
            raise Invalid(f'key: "{key}" contains invalid item "{type(val).__name__}": unable to convert from type "{val}" to integer')

    if not isinstance(val, int):
        raise Invalid(f'key: "{key}" contains invalid item "{val}" with type "{type(val).__name__}": not of type int')

    if min_amount is not None and val < min_amount:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": integer is less then {min_amount}')

    if max_amount is not None and val > max_amount:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": integer is less then {max_amount}')

    return val


def float_validation(val, key=None, min_amount=None, max_amount=None, cast=None, *args, **kwargs):
    if cast:
        try:
            val = float(val)
        except (ValueError, TypeError):
            raise Invalid(f'key: "{key}" contains invalid item "{type(val).__name__}": unable to convert from type "{val}" to float')

    if not isinstance(val, float) or math.isnan(val):
        raise Invalid(f'key: "{key}" contains invalid item "{val}" with type "{type(val).__name__}": not of type float')

    if min_amount is not None and val < min_amount:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": float is less then {min_amount}')

    if max_amount is not None and val > max_amount:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": float is less then {max_amount}')

    return val


def list_validation(val, key=None, min_amount=None, max_amount=None, *args, **kwargs):
    if not isinstance(val, list):
        raise Invalid(f'key: "{key}" contains invalid item "{val}" with type "{type(val).__name__}": not of type list')

    if min_amount is not None and len(val) < min_amount:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": contains less then minimal amount of {min_amount}')

    if max_amount is not None and len(val) > max_amount:
        raise Invalid(f'{key} contains invalid item {val}: contains more then maximum amount of {max_amount}')

    return val


def dict_validation(val, key=None, min_amount=None, max_amount=None, key_regex=None, *args, **kwargs):
    if not isinstance(val, dict):
        raise Invalid(f'"{key}": is not a dictionary')

    if min_amount is not None and len(val) < min_amount:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": {len(val)} contains less then minimal amount of {min_amount}')

    if max_amount is not None and len(val) > max_amount:
        raise Invalid(f'key: "{key}" contains invalid item "{val}": {len(val)} contains more then maximum amount of {max_amount}')

    if key_regex is not None and not all(bool(re.match(key_regex, str(i))) for i in val.keys()):
        for i in val.keys():
            if not re.match(key_regex, str(i)):
                failed = str(i)
                break
        raise Invalid(f'{key}: has dictionary key "{failed}" that does not adhere to regex {key_regex}')

    return val


def uuid_validation(val, key=None, *args, **kwargs):
    try:
        _ = UUID(val, version=4)
    except (ValueError, AttributeError, TypeError):
        raise Invalid('key: "{key}" contains invalid item "{val}": invalid UUID4')

    return val


types = {
    'int': int_validation,
    'str': str_validation,
    'float': float_validation,
    'uuid': uuid_validation,
    'list': list_validation,
    'list_dicts': list_validation,
    'dict': dict_validation,
    'aso_array': dict_validation,
}



import os
import sys
import uuid
import unittest

from deepdiff import DeepDiff as ddiff

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))
import maat

from maat import validate, scale
from maat.exceptions import Invalid

from maat.validations import uuid_validation
from maat.validations import int_validation, str_validation, float_validation, list_validation, dict_validation


class TestNewSyntaxValidation(unittest.TestCase):

    def setUp(self):
        self.test_input = {
            'id': 23,
            'name': 'John Doe',
            'type': 'banana',
        }
        self.test_validation = {
            #'id': {'type': 'int', 'args': {'min_amount': 1}}, old syntax for reference
            'id': {'type': 'int', 'min_amount': 1},
            #'name': {'type': 'str', 'args': {'min_length': 1, 'max_length': 35, 'regex': r'(\w+ )(\w+)'}}
            'name': {'type': 'str', 'min_length': 1, 'max_length': 35, 'regex': r'(\w+ )(\w+)'},
            #'type': {'type': 'str', 'args': {'choices': ['apple', 'banana', 'citrus']}}
            'type': {'type': 'str', 'choices': ['apple', 'banana', 'citrus']}
        }

    def test_validation(self):
        """Happy path test"""
        validated_items = scale(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)

        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

        validated_items = validate(self.test_input, self.test_validation)
        difference = ddiff(validated_items, self.test_input)

        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validation_new_syntax_str_cast(self):
        test_input = {'id': 23}
        expected = {'id': '23'}
        counter_dict = {'id': {'type': 'str', 'cast': True}}
        result = maat.scale(test_input, counter_dict)
        self.assertEqual(expected, result)
        result = maat.validate(test_input, counter_dict)
        self.assertEqual(expected, result)


if __name__ == '__main__':
    unittest.main()



from .maat import scale, Invalid


def protected(validation_dic, fail_is_none=False, custom_exception=None):
    def decorator(func):
        def wrapper(*args, **kwargs):
            if len(args) > 0:
                kwargs.update(zip(func.__code__.co_varnames, args))
            try:
                cleaned_kwargs = scale(kwargs, validation_dic)
            except Invalid:
                if fail_is_none:
                    return None
                elif custom_exception is not None:
                    raise custom_exception
                else:
                    raise
            return func(**cleaned_kwargs)
        return wrapper
    return decorator



import io

from setuptools import setup

with io.open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='Maat',
    author='Melvin Bijman',
    author_email='bijman.m.m@gmail.com',
    version='3.0.8',
    license='MIT',

    py_modules=['maat'],
    packages=['maat'],

    platforms=['any'],

    description='Validate like a Maat',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/Attumm/Maat',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9', 
        'Programming Language :: Python :: 3.10', 
        'Programming Language :: Python :: 3.11', 
        'Programming Language :: Python :: 3.12', 
    ],
)



import os
import sys
import uuid
import unittest

from hypothesis import given
import hypothesis.strategies as st

from deepdiff import DeepDiff as ddiff

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))

from maat import validate

from maat import protected
from maat.exceptions import Invalid


class TestValidations(unittest.TestCase):

    def test_validation_str(self):
        """Happy path test"""
        test_input = {
            "string": "John Doe",
        }
        test_validation = {
            "string": {
                "type": "str",
                "cast": True,
                "min_length": 1,
                "max_length": 12,
                "regex": r"(\w+ )(\w+)",
                "choices": ["John Doe", "Jane Doe"],
            }   
        }
        expected = test_input.copy()

        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validation_int(self):
        """Happy path test"""
        test_input = {
            "int": 33,
        }
        test_validation = {
            "int": {
                "type": "int",
                "cast": True,
                "min_amount": 1,
                "max_amount": 150,
            }
        }
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, test_input)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validation_float(self):
        """Happy path test"""
        test_input = {
            "float": 33.0,
        }
        test_validation = {
            "float": {
                "type": "float",
                "cast": True,
                "min_amount": 1,
                "max_amount": 150,
            }
        }
        expected = test_input.copy()

        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validation_list(self):
        """Happy path test"""
        test_input = {
            "list": ["foo", "bar"],
        }
        test_validation = {
            "list": {
                "type": "list",
                "min_amount": 1,
                "max_amount": 5,
            }
        }
        expected = test_input.copy()

        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validation_dict(self):
        """Happy path test"""
        test_input = {
            "dict": {"foo": "bar", "bar": "foo"},
        }
        test_validation = {
            "dict": {
                "type": "dict",
                "min_amount": 1,
                "max_amount": 2,
                "key_regex": r"(\w+)",
            }
        }
        expected = test_input.copy()

        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validations_combined(self):
        """Copy Past this validation as starting point"""
        test_input = {
            "string": "John Doe",
            "int   ": 33,
            "float ": 33.0,
            "list  ": ["foo", "bar"],
            "dict  ": {"foo": "bar", "bar": "foo"},
        }
        test_validation = {
            "string": {"type": "str", "cast": True, "min_length": 1, "max_length": 12, "regex": r"(\w+ )(\w+)", "choices": ["John Doe", "Jane Doe"]},   
            "int   ": {"type": "int", "cast": True, "min_amount": 1, "max_amount": 150},
            "float ": {"type": "float", "cast": True, "min_amount": 1, "max_amount": 150},
            "list  ": {"type": "list", "min_amount": 1, "max_amount": 5},
            "dict  ": {"type": "dict", "min_amount": 1, "max_amount": 2, "key_regex": r"(\w+)"},
        }
        validated_items = validate(test_input, test_validation)
        expected = test_input.copy()

        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validations_use_default(self):
        test_input = {}
        test_validation = {
            "int   ": {"type": "int", "default": 42},
        }
        expected = {
            "int   ": 42,
        }
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validations_use_optional(self):
        test_input = {}
        test_validation = {
            "float   ": {"type": "float", "optional": True},
        }
        expected = {}
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validations_use_nullable(self):
        test_input = {
            "str   ": None
        }
        test_validation = {
                "str   ": {"type": "str", "nullable": True},
        }
        expected = {
            "str   ": None
        }
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})


    def test_validations_combined_use_nullable_optional_default(self):
        test_input = {
            "str   ": None
        }
        test_validation = {
                "int   ": {"type": "int", "min_amount": 1, "default": 42},
                "float ": {"type": "float", "optional": True},
                "str   ": {"type": "str", "nullable": True},
        }
        expected = {
            "int   ": 42,
            "str   ": None
        }
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validations_nested(self):
        test_input = {
            "foo": {
                "foo_bar": "John Doe Street",
                "foo_baz": 123,
            }
        }
        test_validation = {
            "foo": {"type": "dict", "key_regex": r"\w+", "nested": {
                "foo_bar": {"type": "str", "min_length": 5, "max_length": 99},
                "foo_baz": {"type": "int", "min_amount": 1},
                }
            }
        }
        expected = test_input
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

    def test_validate_nested_list_dict_nested(self):
        test_input = {
            'foobar': [
                {'name': 'John Doe', 'points': 22},
                {'name': 'Jane Doe', 'points': 23},
                {'name': 'willo wanka', 'points': 42},
            ]
        }
        test_validation = {
            'foobar': {'type': 'list_dicts',  'nested': {
                    'name': {'type': 'str'},
                    'points': {'type': 'int'},
                }
            }
        }
        expected = test_input
        validated_items = validate(test_input, test_validation)
        difference = ddiff(validated_items, expected)
        # if the differ finds no difference a empty dictionary is returned
        self.assertEqual(difference, {})

if __name__ == "__main__":
    unittest.main()



from .maat import scale, validate # noqa
from .exceptions import Invalid

from .validations import types
from .validations import int_validation, str_validation, float_validation, list_validation, dict_validation

from .transformations import registered_transformation

from .extras import protected
from .version import VERSION as version

__all__ = ['scale', 'protected', 'Invalid', 'types', 'registered_transformation',
           'int_validation', 'str_validation', 'float_validation', 'list_validation', 'dict_validation', 'uuid_validation'
           'validate', 'version']
